/*
 * Copyright (c) 2011 ICM Uniwersytet Warszawski All rights reserved.
 * See LICENCE file for licencing information.
 *
 * Created on 16-05-2011
 * Author: K. Benedyczak <golbi@mat.umk.pl>
 */
package eu.unicore.ucc.ccm;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.cert.CertificateException;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import org.apache.commons.configuration.ConfigurationException;
import org.apache.commons.configuration.PropertiesConfiguration;
import org.apache.commons.io.FileUtils;

import de.fzj.unicore.ucc.Constants;
import de.fzj.unicore.ucc.MessageWriter;
import eu.unicore.security.canl.TruststoreProperties;
import eu.unicore.util.Log;
import eu.unicore.util.configuration.FilePropertiesHelper;

import xmlbeans.eu.unicore.ccm.internal.ClientProfileDocument.ClientProfile;
import xmlbeans.eu.unicore.ccm.internal.PropertyExtType;
import xmlbeans.eu.unicore.ccm.internal.RegistryType;

/**
 * Updates configuration in properties format. Currently the following things are done here:
 * <ul>
 * <li> it is ensured that properties files exists. If not it is created.
 * <li> all generic properties from the profile are applied to the configuration file.
 * <li> registries are processed and updated according to profile settings
 * <li> it is ensured that truststore settings are set. If not a default type of truststore (JKS) 
 * is set up and configured.
 * </ul>
 * @author K. Benedyczak
 */
public class PropertiesUpdater
{
	public static String DEFAULT_HOST = "localhost";
	public static String DEFAULT_VSITE = "DEMO-SITE";
	
	private MessageWriter parent;
	private String workingDir;
	private boolean removeDefault;
	
	public PropertiesUpdater(MessageWriter reporter, String workingDir)
	{
		this.parent = reporter;
		this.workingDir = workingDir;
	}
	
	public void processRegistries(ClientProfile profile, String propertiesFile, 
			boolean update) throws CCMException
	{
		removeDefault = profile.getRemoveDefaults();
		
		File propsFile = new File(propertiesFile);
		if (!propsFile.exists())
		{
			parent.message("Properties file " + propertiesFile + " does not exist. " +
					"It will be created.");
			try
			{
				File parent = propsFile.getParentFile();
				if (parent != null)
					parent.mkdirs();
				propsFile.createNewFile();
			} catch (IOException e)
			{
				throw new CCMException("Can't create an empty properties file", e);
			}
		}
		
		if (update)
			CCMAction.backupFile(propertiesFile, parent);
		else
		{
			try
			{
				FileUtils.copyFileToDirectory(new File(propertiesFile), 
						new File(workingDir));
				propertiesFile = workingDir + File.separator + 
					new File(propertiesFile).getName();
			} catch (IOException e)
			{
				throw new CCMException("Can't copy the " +
						"original configuration", e);
			}
		}

		
		
		try
		{
			if (profile.isSetClientProperties())
				applyGenericProperties(propertiesFile, profile.getClientProperties().getPropertyArray());			
			
			RegistryType registries[] = profile.getRegistryArray();
			processRegistries(propertiesFile, registries);
			
			setupTruststoreSettings(propertiesFile);
		} catch (ConfigurationException e)
		{
			throw new CCMException("Can't update the " +
					" configuration", e);
		} catch (MalformedURLException e)
		{
			throw new CCMException("Can't update the configuration as " +
					"one of registry URLs was invalid", e);
		} catch (IOException e)
		{
			throw new CCMException("Can't update the configuration as " +
					"there is a I/O error", e);
		} catch (KeyStoreException e)
		{
			throw new CCMException("Can't update the configuration as " +
					"there is a problem with setting up an empty truststore", e);
		} catch (NoSuchAlgorithmException e)
		{
			throw new CCMException("Can't update the configuration as " +
					"there is a problem with setting up an empty truststore", e);
		} catch (CertificateException e)
		{
			throw new CCMException("Can't update the configuration as " +
					"there is a problem with setting up an empty truststore", e);
		}
	}
	
	private void applyGenericProperties(String file, PropertyExtType[] props) throws ConfigurationException, IOException
	{
		if (props == null || props.length == 0)
			return;
		PropertiesFileHelper propsHelper = new PropertiesFileHelper(file);
		PropertiesConfiguration config = propsHelper.getPropertiesConfiguration();
		
		for (PropertyExtType property: props)
		{
			String propertyName = property.getName();
			if (config.containsKey(propertyName))
			{
				if (property.getOverwrite())
				{
					parent.verbose("Overwriting an existing value of property " + propertyName + 
							" (" + config.getString(propertyName) + ") to: " + property.getStringValue());
					config.setProperty(property.getName(), property.getStringValue());
					
				} else
				{
					parent.verbose("Ignoring profile value of property " + propertyName + 
							" (" + property.getStringValue() + 
							") and leaving the current value: " + config.getString(propertyName));
				}
			} else
			{
				parent.verbose("Adding a property " + propertyName + " with value: " + 
						property.getStringValue());
				config.setProperty(property.getName(), property.getStringValue());
			}
		}
		
		propsHelper.close();
	}
	
	private void setupTruststoreSettings(String file) throws ConfigurationException, IOException, CCMException, KeyStoreException, NoSuchAlgorithmException, CertificateException
	{
		PropertiesFileHelper propsHelper = new PropertiesFileHelper(file);
		PropertiesConfiguration config = propsHelper.getPropertiesConfiguration();
		String type = config.getString(TruststoreProperties.DEFAULT_PREFIX+TruststoreProperties.PROP_TYPE);
		if (type == null)
		{
			type = TruststoreProperties.TruststoreType.directory.toString();
			config.setProperty(TruststoreProperties.DEFAULT_PREFIX+TruststoreProperties.PROP_TYPE, 
					type);
			propsHelper.close();
			propsHelper = new PropertiesFileHelper(file);
			config = propsHelper.getPropertiesConfiguration();
		}
		
		FilePropertiesHelper trustHelper = new FilePropertiesHelper(TruststoreProperties.DEFAULT_PREFIX, 
				file, TruststoreProperties.META, Log.getLogger(Log.CONFIGURATION, getClass()));
		if (type.equals(TruststoreProperties.TruststoreType.keystore.toString()))
		{
			String ksType = trustHelper.getValue(TruststoreProperties.PROP_KS_TYPE);
			if (ksType == null)
				ksType = "jks";
			else
				ksType = ksType.toLowerCase();
			
			String ksPath = trustHelper.getValue(TruststoreProperties.PROP_KS_PATH);
			if (ksPath == null)
			{
				ksPath = workingDir + File.separator + "uccTruststore."+ksType;
				config.setProperty(TruststoreProperties.DEFAULT_PREFIX+TruststoreProperties.PROP_KS_PATH, ksPath);
				File ksFile = new File(ksPath);
				if (!ksFile.exists())
				{
					KeyStore ks = KeyStore.getInstance(ksType);
					String tsPasswd = "the!client";
					ks.store(new FileOutputStream(ksFile), tsPasswd.toCharArray());
					config.setProperty(TruststoreProperties.DEFAULT_PREFIX+
							TruststoreProperties.PROP_KS_PASSWORD, tsPasswd);
				} 
			}
		} else if (type.equals(TruststoreProperties.TruststoreType.directory.toString()))
		{
			List<String> paths = trustHelper.getListOfValues(TruststoreProperties.PROP_DIRECTORY_LOCATIONS);
			if (paths.size() == 0)
			{
				String p = workingDir + File.separator + "uccTruststore.d";
				config.setProperty(TruststoreProperties.DEFAULT_PREFIX+TruststoreProperties.PROP_DIRECTORY_LOCATIONS,
						p+File.separator+"*.*");
				File dirFile = new File(p);
				if (!dirFile.exists())
					dirFile.mkdirs();
				else if (!dirFile.isDirectory())
					throw new CCMException("The default path which is used for directory " +
							"type truststore is occupied by a file: " + dirFile);
			}
			
		} else if (type.equals(TruststoreProperties.TruststoreType.openssl.toString()))
		{
			//do nothing - Openssl truststore is unsupported by CCM, will be logged by trust updater.
		} else
			throw new CCMException("Unknown type of truststore is configured: " + type);
		
		propsHelper.close();
	}
	
	private void processRegistries(String file, RegistryType []registries) 
		throws ConfigurationException, IOException
	{
		if (registries == null || registries.length == 0)
		{
			parent.verbose("No registries defined in the configuration profile, " +
					"skipping an update of properties.");
			return;
		}

		Map<URL, RegistryType> newRegs = new LinkedHashMap<URL, RegistryType>();
		for (RegistryType reg: registries)
		{
			URL u = new URL(reg.getAddress());
			newRegs.put(u, reg);
		}
	
		PropertiesFileHelper propsHelper = new PropertiesFileHelper(file);
		PropertiesConfiguration config = propsHelper.getPropertiesConfiguration();
		
		String registryURL = config.getString(Constants.OPT_REGISTRY_LONG);
		String[] urls = registryURL != null ? registryURL.split("[, ]") : new String[0];
		
		StringBuilder newRegistries = new StringBuilder();
		boolean first = true;
		for (String existing: urls)
		{
			if (existing.trim().equals(""))
				continue;
			URL rURL = new URL(existing);
			if (removeDefault)
			{
				if (rURL.getHost().equals(DEFAULT_HOST) && 
						rURL.getPath().contains(DEFAULT_VSITE))
				{
					parent.verbose("Removing the default registry: " + existing);
					continue;
				}
			}
			if (newRegs.containsKey(rURL))
			{
				parent.verbose("Registry " + rURL + " already present.");
				newRegs.remove(rURL);
			}
			if (!first)
				newRegistries.append(" ");				
			newRegistries.append(existing);
			first = false;
		}
		
		Iterator<URL> newOnes = newRegs.keySet().iterator();
		while (newOnes.hasNext())
		{
			if (!first)
				newRegistries.append(" ");
			String toAdd = newOnes.next().toString();
			newRegistries.append(toAdd);
			first = false;
			parent.message("Added a new registry: " + toAdd);
		}
		
		config.setProperty(Constants.OPT_REGISTRY_LONG, newRegistries.toString());
		
		propsHelper.close();
	}	
}











