package de.fzj.unicore.ucc.actions;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Calendar;

import org.apache.commons.cli.OptionBuilder;
import org.unigrids.services.atomic.types.GridFileType;

import de.fzj.unicore.uas.client.StorageClient;

/**
 * Use server-side Find operation to find files on a storage
 * 
 * @author schuller
 */
public class Find extends SMSOperation {
	
	protected boolean detailed;
	
	protected boolean recurse;
	
	protected String base;
	
	protected String name;
	
	protected DateFormat dateFormatter=new SimpleDateFormat("yyyy-MM-dd HH:mm");
	
	protected StorageClient sms;

	//for unit testing
	static int numLastResults;
	
	public String getName(){
		return "find";
	}
	
	@Override
	@SuppressWarnings("all")
	protected void createOptions() {
		super.createOptions();

		getOptions().addOption(OptionBuilder.withLongOpt(OPT_NAME_LONG)
				.withDescription("name match")
				.isRequired(false)
				.hasArg()
				.create(OPT_NAME)
			);

		getOptions().addOption(OptionBuilder.withLongOpt(OPT_DETAILED_LONG)
				.withDescription("detailed listing")
				.isRequired(false)
				.create(OPT_DETAILED)
			);

		getOptions().addOption(OptionBuilder.withLongOpt(OPT_RECURSIVE_LONG)
				.withDescription("recurse into subdirectories")
				.isRequired(false)
				.create(OPT_RECURSIVE)
			);

	}

	@Override
	public void process() {
		super.process();
		detailed=getBooleanOption(OPT_DETAILED_LONG, OPT_DETAILED);
		if(detailed)verbose("Detailed listing.");
		recurse=getBooleanOption(OPT_RECURSIVE_LONG, OPT_RECURSIVE);
		if(recurse)verbose("Listing subdirectories.");
		name=getOption(OPT_NAME_LONG, OPT_NAME);
		doProcess();
	}

	
	protected void doProcess(){
		numLastResults=0;
		
		if(getCommandLine().getArgs().length<2){
			message("Please provide a base path and a storage address ");
			printUsage();
			endProcessing(ERROR_CLIENT);
		}
		String address=getCommandLine().getArgs()[1];
		verbose("Using address <"+address+">");
		
		sms = getStorageClient(address);
		
		if(getCommandLine().getArgs().length>2) {
			base=getCommandLine().getArgs()[2];
		}
		else {
			base=getPathAtStorage(address);
		}
		
		doListing(base);
		
	}
	
	protected void doListing(String path){
		try{
			GridFileType[] listing=sms.find(base,recurse,name,false,null,null);
			for(GridFileType f: listing){
				message(detailed? detailedListing(f):normalListing(f));
				numLastResults++;
			}
		}catch(Exception e){
			error("Error listing path: "+path,e);
			endProcessing(ERROR);
		}
	}
	
	protected String normalListing(GridFileType f){
		return f.getPath();
	}
	
	public String detailedListing(GridFileType f){
		String r=" ";
		String w=" ";
		String x=" ";
		String d=f.getIsDirectory()?"d":"-";
		
		if(f.getPermissions()!=null){
			r=f.getPermissions().getReadable()?"r":"-";
			w=f.getPermissions().getWritable()?"w":"-";
			x=f.getPermissions().getExecutable()?"x":"-";
		}
		Calendar lastMod=f.getLastModified();
		String lastModification=lastMod!=null?format(lastMod):"";
		return String.format("%1$1s%2$1s%3$1s%4$1s %5$14d %6$s %7$-30s",
                d, r, w, x, f.getSize(), lastModification, f.getPath());
}
	
	@Override
	public String getDescription(){
		return "find files on storages";
	}
	
	@Override
	public String getSynopsis(){
		return "find files on storages." +
				"The storage can be given by EPR or using the unicore6:// URL notation.";
	}
	
	@Override
	public String getArgumentList(){
		return "<EPR or u6://...> <base>";
	}
	
	public String format(Calendar c){
		return dateFormatter.format(c.getTime());
	}
	@Override
	public String getCommandGroup(){
		return "Data management";
	}
	
	//for unit testing
	static int getLastNumberOfResults(){
		return numLastResults;
	}
}
