/*
 * Copyright (c) 2011 ICM Uniwersytet Warszawski All rights reserved.
 * See LICENCE file for licencing information.
 *
 * Created on 27-07-2011
 * Author: K. Benedyczak <golbi@mat.umk.pl>
 */
package de.fzj.unicore.ucc.actions;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Calendar;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.OptionBuilder;
import org.json.JSONException;
import org.json.JSONObject;
import org.unigrids.services.atomic.types.ACLEntryType;
import org.unigrids.services.atomic.types.ACLEntryTypeType;
import org.unigrids.services.atomic.types.GridFileType;

import de.fzj.unicore.uas.client.MetadataClient;
import de.fzj.unicore.uas.client.StorageClient;
import de.fzj.unicore.ucc.util.JSONUtil;
import de.fzj.unicore.ucc.util.UnitParser;

/**
 * Gets detailed information about a file.
 * 
 * @author K. Benedyczak
 */
public class GetFileProperties extends SMSOperation {
	
	protected DateFormat dateFormatter = new SimpleDateFormat("yyyy-MM-dd HH:mm");

	protected UnitParser unitParser = UnitParser.getCapacitiesParser(1);
	
	@Override
	@SuppressWarnings("all")
	protected void createOptions() {
		super.createOptions();

		getOptions().addOption(OptionBuilder.withLongOpt(OPT_HUMAN_LONG)
				.withDescription("human-friendly format")
				.isRequired(false)
				.create(OPT_HUMAN)
		);

		getOptions().addOption(OptionBuilder.withLongOpt(OPT_SHOW_META_LONG)
				.withDescription("Show metadata")
				.isRequired(false)
				.create(OPT_SHOW_META)
		);
	}

	@Override
	public String getName() {
		return "stat";
	}

	@Override
	public String getSynopsis() {
		return "Retrieves file status: permissions, ACL, owner and more.";
	}
	
	@Override
	public String getDescription(){
		return "show detailed file information";
	}
	
	@Override
	public String getArgumentList(){
		return "<[Storage-URL#]/path or [u6://SITENAME/[JobId|StorageName]]/path>";
	}
	
	@Override
	public String getCommandGroup(){
		return "Data management";
	}
	
	
	@Override
	public void process() {
		super.process();
		boolean human = getBooleanOption(OPT_HUMAN_LONG, OPT_HUMAN);
		if (human)
			verbose("Human friendly number format.");
		boolean showMetadata = getBooleanOption(OPT_SHOW_META_LONG, OPT_SHOW_META);
		if (showMetadata)
			verbose("Showing metadata.");
		
		CommandLine cmdLine = getCommandLine(); 
		if (cmdLine.getArgs().length != 2) {
			error("Wrong number of arguments", null);
			printUsage();
			endProcessing(ERROR_CLIENT);
		}
		String target = cmdLine.getArgs()[1];
		
		verbose("Getting file properties of " + target);
		
		StorageClient sms = getStorageClient(target);
		
		try{
			
			GridFileType gridFile = sms.listProperties(getPathAtStorage(target));
			String msg = formatStat(gridFile, human, showMetadata);
			message(msg);
		}catch(Exception ex){
			error("Error while getting file properties.",ex);
			endProcessing(ERROR);
		}
	}
	
	private String formatStat(GridFileType f, boolean human, boolean showMetadata) {
		Calendar lastMod = f.getLastModified();
		String lastModification = lastMod != null ? 
				dateFormatter.format(lastMod.getTime()) : "";
		String size = human ? unitParser.getHumanReadable(f.getSize()) :
			String.valueOf(f.getSize());
		String type = f.getIsDirectory() ? "directory" : "normal file";
		
		StringBuilder sb = new StringBuilder();
		sb.append("File: ").append(f.getPath()).append("\n");
		sb.append("Size: ").append(size).append("\n");
		sb.append("Type: ").append(type).append("\n");
		
		if (f.getOwner() != null) 
			sb.append("Owner uid: ").append(f.getOwner()).append("\n");
		if (f.getGroup() != null) 
			sb.append("Owner gid: ").append(f.getGroup()).append("\n");

		if (f.getFilePermissions() != null)
			sb.append("UNIX permissions: ").append(f.getFilePermissions()).append("\n");
		
		if (f.getPermissions() != null) {
			String r = f.getPermissions().getReadable() ? "r" : "-";;
			String w = f.getPermissions().getWritable() ? "w" : "-";
			String x = f.getPermissions().getExecutable() ? "x" : "-";
			sb.append("Current user permissions: ").append(r).append(w).append(x).append("\n");
		}

		if (lastMod != null)
			sb.append("Last modification: ").append(lastModification).append("\n");
		
		if (f.getExtraInformation() != null)
			sb.append("Extra information: ").append(f.getExtraInformation()).append("\n");
		
		if (!f.getIsDirectory() && showMetadata && f.getMetadata() != null) {
			JSONObject o = JSONUtil.asJSON(MetadataClient.asMap(f.getMetadata()));
			try
			{
				sb.append("Metadata: ").append(o.toString(2));
			} catch (JSONException e)
			{
				error("Can't print metadata", e);
			}
		}
		
		if (f.getACL() != null && f.getACL().getEntryArray() != null) {
			ACLEntryType[] acl = f.getACL().getEntryArray();
			if (acl.length > 0) 
				sb.append("ACL:\n");
			for (ACLEntryType ace: acl)
				sb.append("  ").append(formatACE(ace)).append("\n");
		}
		return sb.toString();
	}
	
	private String formatACE(ACLEntryType ace) {
		StringBuilder sb = new StringBuilder();
		if (ace.isSetDefaultACL() && ace.getDefaultACL())
			sb.append("default:");
		
		if (ace.getType().equals(ACLEntryTypeType.USER))
			sb.append("user:");
		else
			sb.append("group:");
		
		if (ace.getSubject() != null)
			sb.append(ace.getSubject());
		sb.append(":");
		
		sb.append(ace.getPermissions());
		return sb.toString();
	}
}





