package de.fzj.unicore.ucc.actions;

import java.util.Calendar;

import org.apache.commons.cli.OptionBuilder;
import org.json.JSONException;
import org.json.JSONObject;
import org.unigrids.services.atomic.types.GridFileType;
import org.unigrids.services.atomic.types.MetadataType;

import de.fzj.unicore.uas.client.MetadataClient;
import de.fzj.unicore.uas.client.StorageClient;
import de.fzj.unicore.ucc.util.JSONUtil;
import de.fzj.unicore.ucc.util.UnitParser;

/**
 * lists a storage
 * 
 * @author schuller
 */
public class LS extends SMSOperation {

	protected boolean detailed;

	protected boolean recurse;

	protected boolean human;

	protected boolean showMetadata;

	protected UnitParser unitParser=UnitParser.getCapacitiesParser(1);

	protected StorageClient sms;

	public String getName(){
		return "ls";
	}

	@Override
	@SuppressWarnings("all")
	protected void createOptions() {
		super.createOptions();

		getOptions().addOption(OptionBuilder.withLongOpt(OPT_DETAILED_LONG)
				.withDescription("detailed listing")
				.isRequired(false)
				.create(OPT_DETAILED)
		);

		getOptions().addOption(OptionBuilder.withLongOpt(OPT_RECURSIVE_LONG)
				.withDescription("recurse into subdirectories")
				.isRequired(false)
				.create(OPT_RECURSIVE)
		);

		getOptions().addOption(OptionBuilder.withLongOpt(OPT_HUMAN_LONG)
				.withDescription("human-friendly format")
				.isRequired(false)
				.create(OPT_HUMAN)
		);

		getOptions().addOption(OptionBuilder.withLongOpt(OPT_SHOW_META_LONG)
				.withDescription("Show metadata")
				.isRequired(false)
				.create(OPT_SHOW_META)
		);
	}

	@Override
	public void process() {
		super.process();
		detailed=getBooleanOption(OPT_DETAILED_LONG, OPT_DETAILED);
		if(detailed)verbose("Detailed listing.");
		recurse=getBooleanOption(OPT_RECURSIVE_LONG, OPT_RECURSIVE);
		if(recurse)verbose("Listing subdirectories.");
		human=getBooleanOption(OPT_HUMAN_LONG, OPT_HUMAN);
		if(human)verbose("Human friendly number format.");
		showMetadata=getBooleanOption(OPT_SHOW_META_LONG, OPT_SHOW_META);
		if(showMetadata)verbose("Showing metadata.");
		doProcess();
	}


	protected void doProcess(){
		if(getCommandLine().getArgs().length<2){
			message("Please provide a storage address!");
			printUsage();
			endProcessing(ERROR_CLIENT);
		}
		String address=getCommandLine().getArgs()[1];
		
		sms = getStorageClient(address);
		String path = getPathAtStorage(address);
		
		if(path==null){
			if(getCommandLine().getArgs().length>2){
				path=getCommandLine().getArgs()[2];
			}	
		}
		if(path==null)path="/";

		doListing(path);

	}

	protected void doListing(String path){
		GridFileType listing=null;
		try{
			listing=sms.listProperties(path);
		}catch(Exception e){
			error("Error getting properties for path: "+path,e);
			endProcessing(ERROR);
		}
		if(listing.getIsDirectory()){
			listDirectory(path);
		}
		else{
			listSingleFile(listing);
		}
	}

	protected void listDirectory(String path){
		try{
			GridFileType[] listing=sms.listDirectory(path);
			for(GridFileType f: listing){
				listSingleFile(f);
				if(f.getIsDirectory()&&recurse){
					doListing(f.getPath());
				}
			}
		}catch(Exception e){
			error("Error listing path: "+path,e);
			endProcessing(ERROR);
		}
	}

	protected void listSingleFile(GridFileType file){
		lastLS=file;
		message(detailed? detailedListing(file):normalListing(file));
		if(!file.getIsDirectory() && showMetadata && file.getMetadata()!=null){
			try{
				printMetadata(file.getMetadata());
			}catch(Exception ex){
				logger.error("Error printing metadata for "+file, ex);
				message("Error! Can't print metadata.");
			}
		}
	}

	protected void printMetadata(MetadataType meta)throws JSONException{
		JSONObject o=JSONUtil.asJSON(MetadataClient.asMap(meta));
		message(o.toString(2));
	}

	protected String normalListing(GridFileType f){
		return f.getPath();
	}

	public String detailedListing(GridFileType f){
		String r=" ";
		String w=" ";
		String x=" ";
		String d=f.getIsDirectory()?"d":"-";

		if(f.getPermissions()!=null){
			r=f.getPermissions().getReadable()?"r":"-";
			w=f.getPermissions().getWritable()?"w":"-";
			x=f.getPermissions().getExecutable()?"x":"-";
		}
		Calendar lastMod=f.getLastModified();
		String lastModification=lastMod!=null?format(lastMod):"";
		String size=human?unitParser.getHumanReadable(f.getSize()):	String.valueOf(f.getSize());
		return String.format("%1$1s%2$1s%3$1s%4$1s %5$14s %6$s %7$-30s",
						d, r, w, x, size, lastModification, f.getPath());
	}

	@Override
	public String getDescription(){
		return "list a storage";
	}

	@Override
	public String getSynopsis(){
		return "Lists files on a storage. " +
		"The storage can be given by EPR or using the unicore6:// URL notation.";
	}

	@Override
	public String getArgumentList(){
		return "<EPR or u6://...> [path]";
	}

	public String format(Calendar c){
		return UnitParser.getSimpleDateFormat().format(c.getTime());
	}
	@Override
	public String getCommandGroup(){
		return "Data management";
	}
	
	//for unit testing
	static GridFileType lastLS;
	
	static GridFileType getLastLS(){
		return lastLS;
	}
}
