/*
 * Copyright (c) 2007, 2008 ICM Uniwersytet Warszawski All rights reserved.
 * See LICENCE file for licencing information.
 *
 * Created on Sep 5, 2011
 * Author: K. Stasiak <karol.m.stasiak@gmail.com>
 */

package de.fzj.unicore.ucc.actions;

import java.net.MalformedURLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import pl.edu.icm.unicore.uvos.api.Attribute;
import pl.edu.icm.unicore.uvos.api.Group;
import pl.edu.icm.unicore.uvos.api.exceptions.AttributeNotFoundException;
import pl.edu.icm.unicore.uvos.wsapi.xmlbeans.SAMLXMLBeansMapper;
import pl.edu.icm.unicore.uvos.wsapi.xmlbeans.UnsupportedSAMLException;
import pl.edu.icm.unicore.uvos.wsclient.samlapi.InvalidSignatureException;
import pl.edu.icm.unicore.uvos.wsclient.samlapi.InvocationException;
import pl.edu.icm.unicore.uvos.wsclient.samlapi.SAMLErrorResponseException;
import pl.edu.icm.unicore.uvos.wsclient.samlapi.SAMLVOQueryClient;
import pl.edu.icm.unicore.uvos.wsclient.samlapi.SecuritySetupException;
import de.fzj.unicore.ucc.helpers.AttributeAssertionFetcher;
import eu.unicore.samly2.assertion.Assertion;
import eu.unicore.samly2.exceptions.SAMLParseException;

/**
 * Action for retrieving and displaying lists of attributes.
 * @author Karol Stasiak
 * @author K. Benedyczak
 */
public class ListAttributes extends ActionBase{
	@Override
	public String getName() {
		return "list-attributes";
	}

	@Override
	@SuppressWarnings("all")
	protected void createOptions() {
		super.createOptions();
	}
	
	@Override
	public String getDescription() {
		return "list VO attributes of the user";
	}

	@Override
	public String getCommandGroup(){
		return "Security";
	}

	@Override
	public String getSynopsis() {
		return "List ALL VO attributes. "
			+ "If given a VO URL with the option -"+OPT_VO+", the assertion is fetched from the VO server first. "
			+ "If given a file name with the option -"+OPT_ATTRIBUTEASSERTION+", the assertion is loaded from that file. "
			+ "All attributes in the assertion are displayed in human readable table. The table also indicates" 
			+ " whether the attribute would be included or not when using the scope and filter settings, " +
			"which were provided to this command.";
	}

	@Override
	protected boolean skipConnectingToRegistry(){
		return
			null != getOption(OPT_ATTRIBUTEASSERTION_LONG, OPT_ATTRIBUTEASSERTION)
			 ||
			! "auto".equals(getOption(OPT_VO_LONG, OPT_VO));
	}

	/**
	 * The minimum width of output table.
	 */
	private static final int MINIMUM_TABLE_WIDTH = 63;

	public void prettyPrint(List<Attribute> la){
		ArrayList<String> scopeColumn = new ArrayList<String>();
		ArrayList<String> nameColumn = new ArrayList<String>();
		ArrayList<String> valueColumn = new ArrayList<String>();
		ArrayList<String> filteredColumn = new ArrayList<String>();
		ArrayList<String> descriptions = new ArrayList<String>();
		scopeColumn.add("Scope");
		nameColumn.add("Name");
		valueColumn.add("Value");
		filteredColumn.add("F");
		Collections.sort(la, ATTRIBUTE_COMPARATOR);
		List<Attribute> filteredByName = attributeFilter.filter(la);
		String groupS = getOption(OPT_VOGROUP_LONG, OPT_VOGROUP, null);
		Group group = null;
		if (groupS != null)
			group = new Group(groupS);
		String latestName="";
		for(Attribute attr: la){
			List<String> attrValue = attr.getValues();
			String attrScope = attr.getScope();
			String attrName = attr.getName();
			if(equal(latestName,attrName)){
				attrName="";
			}
			else {
				//horizontal line
				scopeColumn.add(null);
				nameColumn.add(null);
				valueColumn.add(null);
				filteredColumn.add(null);

				latestName=attrName;

				String description = attr.getType().getDescription();
				if(description==null) description = attr.getType().getShortDescription();
				if(description!=null && !"".equals(description)){
					// rows with name starting with "<D>" are considered placeholders for attribute descriptions
					// description ID are stored as strings in value column and refer to 'descriptions' ArrayList
					nameColumn.add("<D>");
					valueColumn.add(Integer.toString(descriptions.size()));
					scopeColumn.add("");
					filteredColumn.add("");
					descriptions.add(description);
	
					scopeColumn.add(null);
					nameColumn.add(null);
					valueColumn.add(null);
					filteredColumn.add(null);
				}
			}
			nameColumn.add(attrName);
			if(attrScope==null) attrScope="<GLOBAL>";
			scopeColumn.add(attrScope);
			boolean scopeFiltered = false;
			if (group != null)
				scopeFiltered = !SAMLVOQueryClient.isInScope(attr, group, true, groupS);
			
			
			for(int cc=1; cc<attrValue.size(); cc++){
				nameColumn.add("");
				scopeColumn.add("");
			}
			List<String> filteredValues = Collections.emptyList();
			boolean attributeFullyFiltered = true;
			for(Attribute filteredAttr: filteredByName) {
				if (equal(filteredAttr.getName(),attr.getName()) &&
					equal(filteredAttr.getScope(),attr.getScope())){
					filteredValues=filteredAttr.getValues();
					attributeFullyFiltered = false; 
					break;
				}
			}
			if(attrValue.isEmpty()){
				valueColumn.add("<no values>");
				if (!scopeFiltered && !attributeFullyFiltered)
					filteredColumn.add(" ");
				else if (scopeFiltered && !attributeFullyFiltered)
					filteredColumn.add("S");
				else if (!scopeFiltered && attributeFullyFiltered)
					filteredColumn.add("N");
				else
					filteredColumn.add("B");
			}
			for(String v: attrValue){
				valueColumn.add(v);
				if(filteredValues.contains(v)){
					if (!scopeFiltered)
						filteredColumn.add(" ");
					else
						filteredColumn.add("S");
				} else {
					if (!scopeFiltered)
						filteredColumn.add("N");
					else
						filteredColumn.add("B");
				}
			}
		}
		//horizontal line
		scopeColumn.add(null);
		nameColumn.add(null);
		valueColumn.add(null);
		filteredColumn.add(null);
		
		
		
		normalize(scopeColumn,0);
		normalize(nameColumn,0);
		normalize(valueColumn,0);
		normalize(filteredColumn,
				MINIMUM_TABLE_WIDTH
				-nameColumn.get(0).length()
				-scopeColumn.get(0).length()
				-valueColumn.get(0).length()
				-3 //for 3 column separators
				);
		for (int i=0; i<nameColumn.size(); i++){
			if(nameColumn.get(i).startsWith("<D>")){
				int width = scopeColumn.get(i).length() + nameColumn.get(i).length() + filteredColumn.get(i).length() + valueColumn.get(i).length() + 3;
				message(wrap(descriptions.get(Integer.parseInt(valueColumn.get(i).trim())), width));
			}
			else {
				message(nameColumn.get(i) + "|" + scopeColumn.get(i) + "|" + valueColumn.get(i) + "|" + filteredColumn.get(i));
			}
		}
		message("");
		message("The last column shows whether the attribute would be excluded by the current " +
				"filtering settings, which were:");
		message("- filtering by group scope: " + ((groupS == null) ? "NONE" : groupS));
		message("- name/value filters: " +  attributeFilter.toString());
		message("Legend: S - excluded by scope, N - excluded by name/value, B - excluded by both scope and name/value");
	}
	
	@Override
	public void process() {
		super.process();

		if((voURL==null || voURL.isEmpty())&&(attrAssertionFilename==null || attrAssertionFilename.isEmpty())){
			message("Cannot list attributes without attribute source. Set VO server or use file contatining an assertion.");
			endProcessing(ERROR);
		}
		
		try {
			verbose("Connecting to VO " + voURL);
			AttributeAssertionFetcher fetcher = new AttributeAssertionFetcher(securityProperties, voURL);
			Assertion fullAssertion = fetcher.fetchAssertion(null, null); 
			List<Attribute> la =  SAMLXMLBeansMapper.convertSAMLAttributes2API(fullAssertion);
			prettyPrint(la);
		} catch (SAMLParseException ex) {
			error("Malformed SAML response", ex);
			endProcessing(ERROR);
		} catch (UnsupportedSAMLException ex) {
			error("Received SAML response contains unsupported SAML features", ex);
			endProcessing(ERROR);
		} catch (InvalidSignatureException ex) {
			error("Error in processing SAML signature", ex);
			endProcessing(ERROR);
		} catch (InvocationException ex) {
			error("Error communicating with remote service", ex);
			endProcessing(ERROR);
		} catch (SAMLErrorResponseException ex) {
			error("Received SAML error: " + ex.toString(), ex);
			endProcessing(ERROR);
		} catch (SecuritySetupException ex) {
			error("Invalid configuration of security: " + ex.getMessage(), ex);
			endProcessing(ERROR);
		} catch (MalformedURLException ex) {
			error("Invalid syntax of the VO URL: " + voURL, ex);
			endProcessing(ERROR);
		} catch (AttributeNotFoundException e)
		{
			error("Shouldn't happen - BUG: implementation filtered all attributes, " +
					"while should filter none.", e);
			endProcessing(ERROR);
		}
	}

	@Override
	protected void fetchAssertionFromServer() {
		//empty, because we want to list all our attributes
	}
	
	/**
	 * Pads each non-null element in the arraylist with spaces
	 * to the length of the longest element in the arraylist,
	 * but not less than the minimal width.
	 * Replaces null elements in the arraylist with series of dashes
	 * of the same length.
	 * The arraylist is modified in place.
	 * @param list the list of strings
	 * @param minWidth minimum length of strings after the normalization
	 */
	private static void normalize(ArrayList<String> list, int minWidth){
		int max=minWidth;
		for(String s:list){
			if(s!=null && s.length()>max) max=s.length();
		}
		char[] tmp=new char[max];
		Arrays.fill(tmp, ' ');
		String spaces=new String(tmp);
		Arrays.fill(tmp, '-');
		String minuses=new String(tmp);
		for(int i=0; i<list.size(); i++){
			if(list.get(i)==null){
				list.set(i,minuses);
			}
			else {
				int l=list.get(i).length();
				if(l<max){
					list.set(i,list.get(i)+spaces.substring(0, max-l));
				}
			}
		}
	}

	private static boolean equal(String s1, String s2){
		return (s1==null && s2==null)||(s1!=null && s1.equals(s2));
	}

	public static final Comparator<Attribute> ATTRIBUTE_COMPARATOR
			= new Comparator<Attribute>(){
		@Override
		public int compare(Attribute t, Attribute t1) {
			int n = t.getName().compareTo(t1.getName());
			if (n!=0) return n;
			if(t.getScope()==null && t1.getScope()!=null){
				return -1;
			}
			if(t.getScope()!=null && t1.getScope()==null){
				return +1;
			}
			if(t.getScope()==null && t1.getScope()==null){
				return 0;
			}
			return t.getScope().compareTo(t1.getScope());
		}
	};
	
	
	private String wrap(String orig, int width){
		return orig;
	}
}
