package de.fzj.unicore.ucc.helpers;

import java.util.concurrent.Callable;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.atomic.AtomicInteger;

import org.apache.log4j.Logger;
import org.w3.x2005.x08.addressing.EndpointReferenceType;

import de.fzj.unicore.uas.client.BaseUASClient;
import de.fzj.unicore.ucc.InaccessibleServiceException;
import de.fzj.unicore.ucc.MessageWriter;
import de.fzj.unicore.wsrflite.xmlbeans.client.IRegistryQuery;
import eu.unicore.util.Log;
import eu.unicore.util.httpclient.HttpUtils;
import eu.unicore.util.httpclient.IClientConfiguration;

/**
 * A target (identified by its EPR) on which some work shall be performed.
 * The actual work to be performed is passed in as a {@link GridTask} using 
 * the {@link #setTask(GridTask)} method.
 * This implements Callable, it can be submitted to an {@link ExecutorService} <br/>.
 *  
 * The intended use is to execute tasks in parallel and asynchronously on a 
 * list of EPRs, for example:
 * <pre>
  for(EndpointReferenceType epr: eprList){
 	GridTaskTarget<T> target=new GridTaskTarget<T>(secProps,registry,messageWriter);
	target.setEpr(epr);
	target.setTask(task);
	ResourcePool.getExecutorService().submit(target);	
  } 
  </pre>
 *
 * @param <V> - the type of result produced by the {@link GridTask}
 *
 * @author schuller
 */
public class GridTaskTarget<V> implements Callable<V> {
	
	private static final Logger logger=Log.getLogger("UCC",GridTaskTarget.class);
	
	protected EndpointReferenceType epr; 
	protected final IClientConfiguration security; 
	protected final IRegistryQuery registry; 
	protected final MessageWriter msg; 
	protected int connectionTimeout=10000;
	protected GridTask<V> task;
	protected BaseUASClient client;
	
	//track running tasks allowing to limit parallelism
	protected static AtomicInteger runningTasks=new AtomicInteger(0);

	public GridTaskTarget(IClientConfiguration securityProperties, IRegistryQuery registry, MessageWriter msg){
		this.security=securityProperties;
		this.registry=registry;
		this.msg=msg;
	}

	public IClientConfiguration getSecurity() {
		return security;
	}

	public IRegistryQuery getRegistry() {
		return registry;
	}

	public MessageWriter getMessageWriter() {
		return msg;
	}

	public int getConnectionTimeout() {
		return connectionTimeout;
	}

	/**
	 * this default implementation of call() will check
	 * whether the service is accessible. If it is, the {@link GridTask}s run() 
	 * method will be invoked to perform the actual task.
	 * 
	 * @return the result of the {@link GridTask}s run() method 
	 * @throws InaccessibleServiceException if the site is not accessible
	 */
	public V call() throws Exception {
		try{
			if(logger.isDebugEnabled())logger.debug("Start task at "+epr.getAddress().getStringValue());
			runningTasks.incrementAndGet();
			if(!isAccessible()){
				String msg="Service "+epr.getAddress().getStringValue()+" is not accessible.";
				logger.debug(msg);
				throw new InaccessibleServiceException(msg);
			}
			return task.run(this);
		}finally{
			runningTasks.decrementAndGet();
		}
	}

	public EndpointReferenceType getEpr() {
		return epr;
	}

	public void setEpr(EndpointReferenceType epr) {
		this.epr = epr;
	}

	public void setTask(GridTask<V> task) {
		this.task = task;
	}

	public void setConnectionTimeOut(int timeout){
		this.connectionTimeout=timeout;
	}
	
	/**
	 * subclasses should override this method to create the correct
	 * client class
	 * 
	 * @return BaseUASClient
	 * @throws Exception
	 */
	public synchronized BaseUASClient getClient()throws Exception{
		if(client==null){
			security.getExtraSettings().setProperty(HttpUtils.CONNECT_TIMEOUT, String.valueOf(connectionTimeout));
			client=new BaseUASClient(epr.getAddress().getStringValue(),epr,security);
		}
		return client;
	}
	
	/**
	 * checks whether the given service is accessible by calling getCurrentTime()
	 */
	protected boolean isAccessible(){
		try{
			getClient().getCurrentTime();
			if(logger.isDebugEnabled())logger.debug("Service at "+epr.getAddress().getStringValue());
			return true;	
		}catch(Exception e){
			if(logger.isDebugEnabled())logger.debug("No access at "+epr.getAddress().getStringValue());
			return false;
		}
	} 
	
	/**
	 * returns the current number of running tasks
	 */
	public static int getRunningTasks() {
		return runningTasks.intValue();
	}

}
