package de.fzj.unicore.ucc.actions;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.io.File;
import java.io.IOException;
import java.util.Calendar;

import org.apache.commons.io.FileUtils;
import org.junit.Test;
import org.oasisOpen.docs.wsrf.rp2.GetResourcePropertyDocumentResponseDocument;
import org.unigrids.services.atomic.types.GridFileType;
import org.unigrids.x2006.x04.services.jms.JobPropertiesDocument;

import de.fzj.unicore.ucc.UCC;
import de.fzj.unicore.ucc.util.EmbeddedTestBase;
import de.fzj.unicore.ucc.util.UnitParser;
import eu.unicore.bugsreporter.annotation.FunctionalTest;
import eu.unicore.bugsreporter.annotation.RegressionTest;

/**
 * Functional tests for the UCC actions. 
 * These run against an embedded UNICORE instance.
 */
public class TestDataMovementActions extends EmbeddedTestBase {

	@FunctionalTest(id="testPutGetFileWithWildcards", description="Tests put-file and get-file including wildcards")
	@Test
	public void test_PutFile_GetFile_WithWildcards()throws IOException{
		connect();
		String storage=createNewUspace();

		String[] args=new String[]{"put-file", "-s","src/test/resources/testfiles/file*",
				"-t","BFT:"+storage+"#/",
				"-c", "src/test/resources/conf/userprefs.embedded",
				"-v"
		};
		UCC.main(args);
		assertEquals(Integer.valueOf(0),UCC.exitCode);

		args=new String[]{"ls","-l","-H", 
				"-c", "src/test/resources/conf/userprefs.embedded",
				"BFT:"+storage+"#/",
		};
		UCC.main(args);
		assertEquals(Integer.valueOf(0),UCC.exitCode);

		File tmpfolder=new File("target/testdata","ucc-"+System.currentTimeMillis());
		if(!tmpfolder.exists()){
			assertTrue("can't create test directory", tmpfolder.mkdirs());
		}
		args=new String[]{"get-file", "-v",
				"-c", "src/test/resources/conf/userprefs.embedded",
				"-s","BFT:"+storage+"#/file*",
				"-t",tmpfolder.getPath(),
				"BFT:"+storage+"#/",
		};
		UCC.main(args);
		assertEquals(Integer.valueOf(0),UCC.exitCode);

		//check that we got all the files
		assertTrue(new File(tmpfolder,"file1").exists());
		assertTrue(new File(tmpfolder,"file2").exists());

	}

	@FunctionalTest(id="testPutGetFileErrorHandling", description="Tests error handling of put-file and get-file.")
	@Test
	public void test_Error_PutFile_GetFile()throws IOException{
		connect();
		String storage=createNewUspace();

		String[] args=new String[]{"put-file", "-s","target/NOSUCHFILE",
				"-t","BFT:"+storage+"#/",
				"-c", "src/test/resources/conf/userprefs.embedded",
				"-v"
		};
		UCC.main(args);
		assertFalse(Integer.valueOf(0).equals(UCC.exitCode));

		//upload a test file
		args=new String[]{"put-file", "-s","src/test/resources/testfiles/file1",
				"-t","BFT:"+storage+"#/",
				"-c", "src/test/resources/conf/userprefs.embedded",
		};
		UCC.main(args);
		assertTrue(Integer.valueOf(0).equals(UCC.exitCode));

		//download to a non-existing target
		String target="target/NOSUCHDIRECTORY";
		args=new String[]{"get-file", "-v",
				"-c", "src/test/resources/conf/userprefs.embedded",
				"-s","BFT:"+storage+"#/file*",
				"-t",target,
		};
		UCC.main(args);
		assertFalse(Integer.valueOf(0).equals(UCC.exitCode));
	}

	@FunctionalTest(id="testLSFind", description="Tests the ls and find commands")
	@Test
	public void test_PutFile_LS_GetFile_Find()throws IOException{
		connect();
		String storage=createNewUspace();

		String[] args=new String[]{"put-file", "-s","src/test/resources/jobs/date.u",
				"-t","BFT:"+storage+"#/test",
				"-c", "src/test/resources/conf/userprefs.embedded",
				"-v"
		};
		UCC.main(args);
		assertEquals(Integer.valueOf(0),UCC.exitCode);

		args=new String[]{"ls","-l","-H", 
				"-c", "src/test/resources/conf/userprefs.embedded",
				"BFT:"+storage+"#/",
		};
		UCC.main(args);
		assertEquals(Integer.valueOf(0),UCC.exitCode);

		File f=File.createTempFile("ucc-test", "data");
		f.deleteOnExit();

		args=new String[]{"get-file", "-t",f.getAbsolutePath(),
				"-s","BFT:"+storage+"#/test",
				"-c", "src/test/resources/conf/userprefs.embedded",
				"-v"
		};
		UCC.main(args);
		assertEquals(Integer.valueOf(0),UCC.exitCode);

		checkFilesOK(new File("src/test/resources/jobs/date.u"), f);

		args=new String[]{"find","-v",
				"-c", "src/test/resources/conf/userprefs.embedded",
				"-N", "/test",
				storage,"/"
		};
		UCC.main(args);
		assertEquals(Integer.valueOf(0),UCC.exitCode);

		//TODO SMS.Find() is buggy! when using "-N test" 3 results are returned
		assertEquals(1,Find.getLastNumberOfResults()); 
	}

	@FunctionalTest(id="testGetFileRawHttp", description="Tests the get-file from a http source")
	@Test
	public void test_GetFile_RawHttpURL()throws IOException{
		connect();

		String[] args=new String[]{"get-file", "-s","http://localhost:65322/services/Registry?wsdl",
				"-t","test-get-file"+System.currentTimeMillis(),
				"-c", "src/test/resources/conf/userprefs.embedded",
				"-v"
		};
		UCC.main(args);
		assertEquals(Integer.valueOf(0),UCC.exitCode);
	}

	@FunctionalTest(id="testFileOperationByteRanges", description="Tests the put/get-file ops using a byte range")
	@Test
	public void test_FileOps_ByteRange()throws IOException{
		connect();
		String storage=createNewUspace();
		String data="0123456789abcdefghijklmnopqrstuvwxyz";
		File t=new File("target","test-upload-"+System.currentTimeMillis());
		FileUtils.writeStringToFile(t, data);
		
		int endByte=3;
		String[] args=new String[]{"put-file", "-t","BFT:"+storage+"#/test",
				"-s", t.getPath(),
				"-c", "src/test/resources/conf/userprefs.embedded",
				"-B", "0-"+endByte,
				"-v"
		};
		UCC.main(args);
		assertEquals(Integer.valueOf(0),UCC.exitCode);
		
		args=new String[]{"ls","-l","-H", 
				"-c", "src/test/resources/conf/userprefs.embedded",
				"BFT:"+storage+"#/test",
		};
		UCC.main(args);
		assertEquals(Integer.valueOf(0),UCC.exitCode);
		
		GridFileType lsResult=LS.getLastLS();
		assertEquals(endByte+1, lsResult.getSize());
		
		//upload fully
		args=new String[]{"put-file", "-t","BFT:"+storage+"#/test2",
				"-s", t.getPath(),
				"-c", "src/test/resources/conf/userprefs.embedded",
				"-v"
		};
		UCC.main(args);
		assertEquals(Integer.valueOf(0),UCC.exitCode);
		
		args=new String[]{"ls","-l","-H", 
				"-c", "src/test/resources/conf/userprefs.embedded",
				"BFT:"+storage+"#/test",
		};
		UCC.main(args);
		assertEquals(Integer.valueOf(0),UCC.exitCode);
		
		//get part of file
		t=new File("target","test-get-file"+System.currentTimeMillis());
		args=new String[]{"get-file", "-s","BFT:"+storage+"#/test2",
				"-t",t.getPath(),
				"-B","1-4",
				"-c", "src/test/resources/conf/userprefs.embedded",
				"-v"
		};
		UCC.main(args);
		assertEquals(Integer.valueOf(0),UCC.exitCode);
		assertEquals(4l,t.length());
		
		String content=FileUtils.readFileToString(t);
		assertEquals("1234",content);
		
		//get-file using raw http
		args=new String[]{"get-file", "-s", "http://localhost:65322/services/Registry?wsdl",
				"-t",t.getPath(),
				"-B","1-4",
				"-c", "src/test/resources/conf/userprefs.embedded",
				"-v"
		};
		UCC.main(args);
		assertEquals(Integer.valueOf(0),UCC.exitCode);
	}

	
	@FunctionalTest(id="testCopyFileStatus", description="Tests the copy-file and copy-file-status commands")
	@Test
	public void test_CopyFile_and_CopyFileStatus()throws IOException{
		connect();
		String storage=createNewUspace();
		String storage2=createNewUspace();

		String[] args=new String[]{"put-file", "-s","src/test/resources/jobs/date.u",
				"-t","BFT:"+storage+"#/test",
				"-c", "src/test/resources/conf/userprefs.embedded",

		};
		UCC.main(args);
		assertEquals(Integer.valueOf(0),UCC.exitCode);

		args=new String[]{"copy-file", 
				"-c", "src/test/resources/conf/userprefs.embedded",
				"-s","BFT:"+storage+"#/test",
				"-t","BFT:"+storage2+"#/test",
		};
		UCC.main(args);
		assertEquals(Integer.valueOf(0),UCC.exitCode);

		File f=File.createTempFile("ucc-test", "data");
		f.deleteOnExit();

		args=new String[]{"get-file", "-t",f.getAbsolutePath(),
				"-s","BFT:"+storage2+"#/test",
				"-c", "src/test/resources/conf/userprefs.embedded",
		};
		UCC.main(args);
		assertEquals(Integer.valueOf(0),UCC.exitCode);

		checkFilesOK(new File("src/test/resources/jobs/date.u"), f);

		//same in async mode
		args=new String[]{"copy-file", 
				"-c", "src/test/resources/conf/userprefs.embedded",
				"-s","BFT:"+storage+"#/test",
				"-t","BFT:"+storage2+"#/test_async",
				"-a","-v"
		};
		UCC.main(args);
		assertEquals(Integer.valueOf(0),UCC.exitCode);

		args=new String[]{"copy-file-status", 
				"-c", "src/test/resources/conf/userprefs.embedded",
				CopyFile.getLastFTAddress(),
		};
		UCC.main(args);
		assertEquals(Integer.valueOf(0),UCC.exitCode);


		File f2=File.createTempFile("ucc-test", "data-2");
		f2.deleteOnExit();
		args=new String[]{"get-file", "-t",f2.getAbsolutePath(),
				"-s","BFT:"+storage2+"#/test",
				"-c", "src/test/resources/conf/userprefs.embedded",
		};
		UCC.main(args);
		assertEquals(Integer.valueOf(0),UCC.exitCode);

		checkFilesOK(new File("src/test/resources/jobs/date.u"), f2);
	}

	@FunctionalTest(id="testListFiletransfers", description="Tests the list-transfers commands")
	@Test
	public void test_ListTransfers()throws IOException{
		connect();
		String storage=createStorage();
		String storage2=createStorage();

		String[] args=new String[]{"put-file", "-s","src/test/resources/jobs/date.u",
				"-t","BFT:"+storage+"#/test",
				"-c", "src/test/resources/conf/userprefs.embedded",

		};
		UCC.main(args);
		assertEquals(Integer.valueOf(0),UCC.exitCode);
		//copy in async mode
		args=new String[]{"copy-file", 
				"-c", "src/test/resources/conf/userprefs.embedded",
				"-s","BFT:"+storage+"#/test",
				"-t","BFT:"+storage2+"#/test_async",
				"-a",
		};
		UCC.main(args);
		assertEquals(Integer.valueOf(0),UCC.exitCode);

		//scheduled in async mode
		Calendar startTime=Calendar.getInstance();
		startTime.add(Calendar.DATE, 1);
		
		args=new String[]{"copy-file", 
				"-c", "src/test/resources/conf/userprefs.embedded",
				"-s","BFT:"+storage+"#/test",
				"-t","BFT:"+storage2+"#/test_async",
				"-a","-S", UnitParser.getISO8601().format(startTime.getTime()),
		};
		UCC.main(args);
		assertEquals(Integer.valueOf(0),UCC.exitCode);

		
		args=new String[]{"list-transfers", 
				"-c", "src/test/resources/conf/userprefs.embedded",
				"-l", 
		};
		UCC.main(args);
		assertEquals(Integer.valueOf(0),UCC.exitCode);
	}

	@RegressionTest(url="https://sourceforge.net/tracker/?func=detail&aid=3406663&group_id=102081&atid=633902")
	@Test
	public void test_Error_CopyFile()throws IOException{
		connect();
		String storage=createNewUspace();
		String storage2=createNewUspace();

		String[] args=new String[]{"copy-file", 
				"-c", "src/test/resources/conf/userprefs.embedded",
				"-s","BFT:"+storage+"#/NOSUCHFILE",
				"-t","BFT:"+storage2+"#/",
		};
		UCC.main(args);
		assertFalse(Integer.valueOf(0)==UCC.exitCode);
	}

	@FunctionalTest(id="testGetDirectory", description="Tests get-file of a directory")
	@Test
	public void test_ExportDirectory()throws IOException{
		connect();
		String storage=createNewUspace();
		//upload files
		String[] args=new String[]{"put-file", "-s","src/test/resources/testfiles/*",
				"-t","BFT:"+storage+"#/",
				"-c", "src/test/resources/conf/userprefs.embedded",
		};
		UCC.main(args);
		assertEquals(Integer.valueOf(0),UCC.exitCode);
		//mkdir
		args=new String[]{"mkdir","BFT:"+storage+"#/sub",
				"-c", "src/test/resources/conf/userprefs.embedded",
		};
		UCC.main(args);
		assertEquals(Integer.valueOf(0),UCC.exitCode);
		//upload more files
		args=new String[]{"put-file", "-s","src/test/resources/testfiles/*",
				"-t","BFT:"+storage+"#/sub/",
				"-c", "src/test/resources/conf/userprefs.embedded",
		};
		UCC.main(args);
		assertEquals(Integer.valueOf(0),UCC.exitCode);

		//and export the complete directory
		File localTarget=new File("target","ucc-export-"+System.currentTimeMillis());
		localTarget.mkdirs();
		args=new String[]{"get-file", "-t",localTarget.getPath(),
				"-s","BFT:"+storage+"#/", 
				"-R",
				"-c", "src/test/resources/conf/userprefs.embedded",
		};
		UCC.main(args);
		assertEquals(Integer.valueOf(0),UCC.exitCode);
		//check subdir is there
		File sub=new File(localTarget,"sub");
		assertTrue(sub.exists() && sub.isDirectory());
		//and that files in subdirectory exist
		assertEquals(3, sub.listFiles().length);
	}


	@FunctionalTest(id="testPutDirectory", description="Tests put-file of a directory")
	@Test
	public void test_ImportDirectory()throws IOException{
		connect();
		String storage=createNewUspace();
		//create a directory with a few files
		String dirname="ucctest-"+System.currentTimeMillis();
		File tmp=new File(System.getProperty("java.io.tmpdir"), dirname);
		assertTrue("Can't create tmpdir",tmp.mkdir());
		FileUtils.writeStringToFile(new File(tmp,"test1"), "test1");
		FileUtils.writeStringToFile(new File(tmp,"test2"), "test2");
		File subdir=new File(tmp, "sub");
		assertTrue("Can't create tmpdir",subdir.mkdir());
		FileUtils.writeStringToFile(new File(subdir,"sub1"), "sub1");

		//upload a complete directory
		String[] args=new String[]{"put-file", "-s",tmp.getAbsolutePath(),
				"-t","BFT:"+storage+"#/",
				"-R",
				"-c", "src/test/resources/conf/userprefs.embedded",
		};
		UCC.main(args);
		assertEquals(Integer.valueOf(0),UCC.exitCode);

		//export the complete directory
		File localTarget=new File("target","ucc-export-"+System.currentTimeMillis());
		localTarget.mkdirs();
		args=new String[]{"get-file", "-t",localTarget.getPath(),
				"-s","BFT:"+storage+"#/", 
				"-R",
				"-c", "src/test/resources/conf/userprefs.embedded",
		};
		UCC.main(args);
		assertEquals(Integer.valueOf(0),UCC.exitCode);
		//check subdir is there
		File sub=new File(localTarget,dirname+"/sub");
		assertTrue(sub.exists() && sub.isDirectory());
		//and that files in subdirectory exist
		assertTrue(sub.listFiles().length==1);
	}

	@FunctionalTest(id="testCatFile", description="Tests the 'cat' command")
	@Test
	public void test_CatFile()throws IOException{
		connect();
		String storage=createNewUspace();
		String[] args=new String[]{"cat",
				"BFT:"+storage+"#/stdout",
				"-c", "src/test/resources/conf/userprefs.embedded",
		};
		UCC.main(args);
		assertEquals(Integer.valueOf(0),UCC.exitCode);
		
		args=new String[]{"cat",
				"BFT:"+storage+"#/stdout",
				"BFT:"+storage+"#/stdout",
				"BFT:"+storage+"#/stdout",
				"-c", "src/test/resources/conf/userprefs.embedded",
		};
		UCC.main(args);
		assertEquals(Integer.valueOf(0),UCC.exitCode);
		
		args=new String[]{"cat",
				"-c", "src/test/resources/conf/userprefs.embedded",
		};
		UCC.main(args);
		assertEquals(Integer.valueOf(1),UCC.exitCode);
		
		args=new String[]{"cat",
				"BFT:"+storage+"#/std*",
				"-c", "src/test/resources/conf/userprefs.embedded",
		};
		UCC.main(args);
		assertEquals(Integer.valueOf(0),UCC.exitCode);
	}
	
	/**
	 * creates a new uspace and returns its URL
	 * @return
	 */
	protected String createNewUspace(){
		String[] args=new String[]{"run",
				"-c", "src/test/resources/conf/userprefs.embedded",
				"src/test/resources/jobs/date.u",
		};
		UCC.main(args);
		assertEquals(Integer.valueOf(0),UCC.exitCode);

		args=new String[]{"wsrf", "getproperties", 
				"-c", "src/test/resources/conf/userprefs.embedded",
				Run.getLastJobAddress(),
		};
		UCC.main(args);
		assertEquals(Integer.valueOf(0),UCC.exitCode);

		try{
			GetResourcePropertyDocumentResponseDocument d=GetResourcePropertyDocumentResponseDocument.Factory.parse(WSRF.getLastProperties());
			JobPropertiesDocument jd= JobPropertiesDocument.Factory.parse(d.getGetResourcePropertyDocumentResponse().newInputStream());
			return jd.getJobProperties().getWorkingDirectoryReference().getAddress().getStringValue();
		}catch(Exception ex){
			ex.printStackTrace();
			fail(); 
		}
		return null;
	}

	protected String createStorage(){
		String[] args=new String[]{"create-storage", "-v", 
				"-c", "src/test/resources/conf/userprefs.embedded",
		};
		UCC.main(args);
		assertEquals(Integer.valueOf(0),UCC.exitCode);
		return CreateStorage.getLastStorageAddress();
	}

}
