package de.fzj.unicore.ucc.util;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.io.File;
import java.util.List;
import java.util.Set;

import org.apache.xmlbeans.XmlObject;
import org.ggf.schemas.jsdl.x2005.x11.jsdl.ApplicationType;
import org.ggf.schemas.jsdl.x2005.x11.jsdl.DataStagingType;
import org.ggf.schemas.jsdl.x2005.x11.jsdl.JobDefinitionDocument;
import org.ggf.schemas.jsdl.x2005.x11.jsdl.OperatingSystemTypeEnumeration;
import org.ggf.schemas.jsdl.x2005.x11.jsdl.ResourcesType;
import org.ggf.schemas.jsdl.x2005.x11.jsdlPosix.EnvironmentType;
import org.ggf.schemas.jsdl.x2005.x11.jsdlPosix.POSIXApplicationDocument;
import org.ggf.schemas.jsdl.x2005.x11.jsdlPosix.POSIXApplicationType;
import org.json.JSONObject;
import org.junit.Test;
import org.unigrids.services.atomic.types.ProtocolType;
import org.unigrids.x2006.x04.services.tss.ApplicationResourceType;
import org.unigrids.x2006.x04.services.tss.TargetSystemPropertiesDocument.TargetSystemProperties;

import de.fzj.unicore.ucc.helpers.DefaultMessageWriter;
import de.fzj.unicore.ucc.runner.ExecutionEnvironmentRequirement;
import de.fzj.unicore.ucc.runner.Requirement;
import de.fzj.unicore.wsrflite.xmlbeans.WSUtilities;

public class TestBuilder{

	//test creating builder from a JSON string
	@Test
	public void testBuildFromJSONString() throws Exception{
		Builder b=new Builder(new File("src/test/resources/jobs/test.u"));
		assertNotNull(b.getJob());
		ApplicationType a=b.getJob().getJobDefinition().getJobDescription().getApplication();
		assertEquals(a.getApplicationName(),"foo");
		assertEquals(a.getApplicationVersion(),"1.0");

		POSIXApplicationType posix=(POSIXApplicationType)WSUtilities.extractAny(a, POSIXApplicationDocument.type.getDocumentElementName())[0];
		assertEquals("-v", posix.getArgumentArray()[0].getStringValue());
		assertEquals("-d", posix.getArgumentArray()[1].getStringValue());
		assertEquals("somebody",posix.getUserName().getStringValue());

		EnvironmentType[] env=posix.getEnvironmentArray();
		assertEquals(2, env.length);
		DataStagingType[] ds=b.getJob().getJobDefinition().getJobDescription().getDataStagingArray();
		assertEquals(2, ds.length);

		assertTrue(ds[0].toString().contains("IgnoreFailure>true")
				|| ds[1].toString().contains("IgnoreFailure>true")
				);
		
		List<FileUploader> imp=b.getImports();
		assertEquals(1,imp.size());
		assertEquals("myfile", imp.get(0).getFrom());
		assertEquals("uspaceFile", imp.get(0).getTo());	

		List<FileDownloader> exp=b.getExports();
		assertEquals(2,exp.size());
		assertEquals("uspaceOutFile", exp.get(0).getFrom());
		assertEquals("localName", exp.get(0).getTo());	
		assertEquals("otherUspaceOutFile", exp.get(1).getFrom());
		assertEquals("otherLocalName", exp.get(1).getTo());

		ResourcesType res=b.getJob().getJobDefinition().getJobDescription().getResources();
		assertEquals(128000000d, res.getIndividualPhysicalMemory().getExactArray(0).getDoubleValue(),0.1);
		assertEquals(32d, res.getIndividualCPUCount().getExactArray(0).getDoubleValue(),0.1);
		assertEquals(3600d, res.getIndividualCPUTime().getExactArray(0).getDoubleValue(),0.1);
		assertEquals(4d, res.getTotalResourceCount().getExactArray(0).getDoubleValue(),0.1);
		assertTrue(res.toString().contains("64738"));

		assertEquals("DEMO-SITE",b.getSite());
		
		String[] proj=b.getJob().getJobDefinition().getJobDescription().getJobIdentification().getJobProjectArray();
		assertNotNull(proj);
		assertEquals(1, proj.length);
		assertEquals("my_project", proj[0]);
		String[] tags=b.getJob().getJobDefinition().getJobDescription().getJobIdentification().getJobAnnotationArray();
		assertNotNull(tags);
		assertEquals(2, tags.length);
	}

	@Test
	public void testBuildFromJSONString2() throws Exception{
		Builder b=new Builder(new File("src/test/resources/jobs/test2.u"));
		assertNotNull(b.getJob());
		ApplicationType a=b.getJob().getJobDefinition().getJobDescription().getApplication();
		assertEquals(a.getApplicationName(),"foo");

		POSIXApplicationType posix=(POSIXApplicationType)WSUtilities.extractAny(a, POSIXApplicationDocument.type.getDocumentElementName())[0];
		EnvironmentType[] env=posix.getEnvironmentArray();
		assertEquals(3, env.length);
		System.out.println(posix);
		assertTrue(posix.toString().contains("bar"));
		assertTrue(posix.toString().contains("spam"));
		assertTrue(posix.toString().contains("debug"));
	}

	
	//test creating the builder using the API
	@Test
	public void testBuildUsingAPI()throws Exception{
		Builder b=new Builder().
		setApplication("foo", "1.0").
		hasArgument("-v").
		hasArgument("-d").
		setEnvironment("SHELL", "/bin/bash").
		withExport("foo", "bar").
		withImport("myfile", "uspaceFile",Mode.overwrite,false);
		
		//set some attributes
		b.setProperty("Lifetime", "4h");
		b.setProperty("Update interval", "10");
		b.setProperty("Preferred protocols", "RBYTEIO SBYTEIO");
		
		assertNotNull(b.getJob());
		ApplicationType a=b.getJob().getJobDefinition().getJobDescription().getApplication();
		assertEquals(a.getApplicationName(),"foo");
		assertEquals(a.getApplicationVersion(),"1.0");

		POSIXApplicationType posix=(POSIXApplicationType)WSUtilities.extractAny(a, POSIXApplicationDocument.type.getDocumentElementName())[0];
		assertEquals("-v", posix.getArgumentArray()[0].getStringValue());
		assertEquals("-d", posix.getArgumentArray()[1].getStringValue());

		List<FileUploader> imp=b.getImports();
		assertEquals(1,imp.size());
		assertEquals("myfile", imp.get(0).getFrom());
		assertEquals("uspaceFile", imp.get(0).getTo());	

		//test attributes
		assertEquals(4*3600,b.getLifetime());
		assertEquals("10",b.getProperty("Update interval"));
		assertEquals(2, b.getPreferredProtocols().length);
		assertEquals(ProtocolType.RBYTEIO, b.getPreferredProtocols()[0]);
		assertEquals(ProtocolType.SBYTEIO, b.getPreferredProtocols()[1]);
	}

	@Test
	public void testRequirements()throws Exception{
		Builder b=new Builder(new File("src/test/resources/jobs/date_withresources.u"));
		assertNotNull(b.getJob());
		TargetSystemProperties ts=TargetSystemProperties.Factory.newInstance();
		OperatingSystemTypeEnumeration.Enum osType=OperatingSystemTypeEnumeration.GNU_HURD;
		ts.addNewOperatingSystem().addNewOperatingSystemType().setOperatingSystemName(osType);
		ApplicationResourceType app=ts.addNewApplicationResource();
		app.setApplicationName("Date");
		app.setApplicationVersion("1.0");

		assertEquals(2, b.getRequirements().size());

		for(Requirement r: b.getRequirements()){
			assertTrue("failed requirement check: "+r.getDescription(),r.isFulfilled(ts));
		}
	}
	
	@Test
	public void testExecutionEnvironment()throws Exception{
		Builder b=new Builder(new File("src/test/resources/jobs/date_with_execution_environment.u"));
		assertNotNull(b.getJob());
		String job=b.getJob().toString();
		assertTrue(job.contains("TIME"));
		assertTrue(job.contains("PRE1"));
		assertTrue(job.contains("PRE2"));
		assertTrue(job.contains("POST1"));
		assertTrue(job.contains("POST2"));
		assertTrue(job.contains("verbose"));
		assertTrue(job.contains("testing"));
		assertTrue(job.contains("TEST"));
		assertTrue(job.contains("touch 456"));
		assertTrue(job.contains("echo 123"));

	}

	@Test
	public void testMinimalExecutionEnvironment()throws Exception{
		Builder b=new Builder(new File("src/test/resources/jobs/date_with_minimal_ee.u"));
		assertNotNull(b.getJob());
		String job=b.getJob().toString();
		assertTrue(job.contains("TIME"));
		Set<Requirement>req=b.getRequirements();
		assertEquals(2,req.size());
		ExecutionEnvironmentRequirement ee=new ExecutionEnvironmentRequirement("TIME", null);
		assertTrue(req.contains(ee));
	}

	@Test
	public void testValidateJSDL()throws Exception{

		String jsdlSpec="<JobDefinition xmlns=\"http://schemas.ggf.org/jsdl/2005/11/jsdl\">"+
		"<JobDescription>   <Application>    <xApplicationName>foo</xApplicationName>  " +
		" </Application>  </JobDescription> </JobDefinition>";
		JobDefinitionDocument jdd=JobDefinitionDocument.Factory.parse(jsdlSpec);

		try{
			assertFalse(Builder.isValidJSDL(jdd, null));
		}catch(Exception ex){
			ex.printStackTrace();
			fail(ex.getMessage());
		}
		jsdlSpec="<JobDefinition xmlns=\"http://schemas.ggf.org/jsdl/2005/11/jsdl\">"+
		"<JobDescription>   <Application>    <ApplicationName>foo</ApplicationName>  " +
		" </Application>  </JobDescription> </JobDefinition>";

		jdd=JobDefinitionDocument.Factory.parse(jsdlSpec);

		try{
			assertTrue(Builder.isValidJSDL(jdd, new DefaultMessageWriter()));
		}catch(Exception ex){
			ex.printStackTrace();
			fail(ex.getMessage());
		}

	}

	//check that the sample job is in fact syntactically valid
	@Test
	public void testPrintSample()throws Exception{
		String sample = Builder.getJobSample();
		JSONObject o=new JSONObject(sample);
		assertEquals("Date",o.get("ApplicationName"));
	}
	
	@Test
	public void testCredential()throws Exception{
		JSONObject cred=new JSONObject();
		cred.put("Username", "foo");
		cred.put("Password", "bar");
		XmlObject co=new Builder().makeCredentials(cred);
		String c=co.toString();
		assertTrue(c.contains("<wsse:Username>foo</wsse:Username>"));
		assertTrue(c.contains("<wsse:Password>bar</wsse:Password>"));
	}
	
	@Test
	public void testCredentialFromJSON()throws Exception{
		String job="{ ApplicationName: date, \n" +
				"Imports: [ \n" +
				"{From: \"scp://test\", To: in, \n" +
				"Credentials: { Username: foo, Password: bar }, " +
				"\n}" +
				"]," +
				"Exports: [ \n" +
				"{To: \"scp://test1\", From: stdout, \n" +
				"Credentials: { Username: ham, Password: spam}, " +
				"\n}" +
				"]," +
				"}";
		Builder b=new Builder(job);
		String jsdl=b.getJob().toString();
		assertTrue(jsdl.contains("<wsse:Username>foo</wsse:Username>"));
		assertTrue(jsdl.contains("<wsse:Password>bar</wsse:Password>"));
		assertTrue(jsdl.contains("<wsse:Username>ham</wsse:Username>"));
		assertTrue(jsdl.contains("<wsse:Password>spam</wsse:Password>"));
	}
	
	@Test
	public void testOS()throws Exception{
		String[] os=new String[]{"linux", "MacOS", "Aix", "WinNT", "UNKnoWn", 
				"freeBSD", "hpux", "netbsd", "irix", "solARIS"};
		Builder Bob=new Builder();//obvious really
		for(String o: os){
			assertNotNull("OS "+o+" not understood", Bob.getOSType(o));
		}
	}
	
}
