package de.fzj.unicore.ucc.bes;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileWriter;
import java.util.Calendar;
import java.util.GregorianCalendar;

import org.apache.commons.cli.OptionBuilder;
import org.ggf.schemas.bes.x2006.x08.besFactory.ActivityDocumentDocument1;
import org.ggf.schemas.bes.x2006.x08.besFactory.ActivityDocumentType;
import org.ggf.schemas.bes.x2006.x08.besFactory.CreateActivityDocument;
import org.ggf.schemas.bes.x2006.x08.besFactory.CreateActivityResponseDocument;
import org.ggf.schemas.bes.x2006.x08.besFactory.CreateActivityType;
import org.ggf.schemas.jsdl.x2005.x11.jsdl.JobDefinitionDocument;
import org.oasisOpen.docs.wsrf.rl2.TerminationTimeDocument.TerminationTime;
import org.w3.x2005.x08.addressing.EndpointReferenceType;

import de.fzj.unicore.bes.faults.InvalidRequestMessageFault;
import de.fzj.unicore.bes.faults.NotAcceptingNewActivitiesFault;
import de.fzj.unicore.bes.faults.UnsupportedFeatureFault;
import de.fzj.unicore.bes.util.ActivityUtils;
import de.fzj.unicore.ucc.util.Builder;

public class RunJob extends BESActionBase implements BESUCCConstants {

	protected String besUrl = null;

	protected String jsdlPath = null;
	
	protected EndpointReferenceType activityEpr;
	
	private ActivityDocumentType activityDocument;

	protected Builder builder;

	/**
	 * synchronous mode
	 */
	protected boolean synchronous;

	/**
	 * do not add job id prefixes to output file names
	 */
	protected boolean brief;

	@Override
	@SuppressWarnings("all")
	protected void createOptions() {
		super.createOptions();

		getOptions().addOption(
				OptionBuilder.withLongOpt(OPT_NOPREFIX_LONG).withDescription(
						"Short output file names").isRequired(false).create(
						OPT_NOPREFIX));

		getOptions().addOption(
				OptionBuilder.withLongOpt(OPT_STDOUT_LONG).withDescription(
						"Name for storing the standard output")
						.withArgName("stdout").hasArg().isRequired(false)
						.create(OPT_STDOUT));

		getOptions().addOption(
				OptionBuilder.withLongOpt(OPT_STDERR_LONG).withDescription(
						"Name for storing the standard error")
						.withArgName("stderr").hasArg().isRequired(false)
						.create(OPT_STDERR));
		getOptions().addOption(
				OptionBuilder.withLongOpt(OPT_JSDL_LONG).withDescription(
						"Treat job file as JSDL document").isRequired(false)
						.create(OPT_JSDL));
	}


	@Override
	public String getName() {
		return "bes-submit-job";
	}

	@Override
	public String getArgumentList() {
		return "[<jobfile>]";
	}

	@Override
	public String getSynopsis() {
		return "Runs a job through OGSA-BES. "
				+ "The job definition is read from <jobfile> or stdin. "
				+ "A descriptor file will be written that can be used later with other ucc commands.";
	}

	@Override
	public String getDescription() {
		return "run a job through UNICORE 6 BES Interface";
	}

	@Override
	public void process() {
		super.process();
						
		setOutputLocation();
	
		if (timing) {
			startTime = System.currentTimeMillis();
		}
		initBuilder();
		if (timing) {
			endTime = System.currentTimeMillis();
			long duration = endTime - startTime;
			message("Time: " + duration + " ms.");
		}
		
		runBes();
	}

	protected void initBuilder() {
		String jobFile = null;
		if (getCommandLine().getArgs().length > 1) {
			jobFile = getCommandLine().getArgs()[1];
			if (getCommandLine().hasOption(OPT_JSDL)) {
				try {
					verbose("Loading JSDL file.");
					builder = new Builder();
					builder.setMessageWriter(this);
					builder.setRegistry(registry);
					JobDefinitionDocument jdd = JobDefinitionDocument.Factory
							.parse(new File(jobFile));
					builder.setJob(jdd);
				} catch (Exception e) {
					error("Can't parse jsdl file <" + jobFile + ">", e);
					endProcessing();
				}
			} else {
				try {
					builder = new Builder(new File(jobFile));
					builder.setMessageWriter(this);
					builder.setRegistry(registry);
					verbose("Read job from <" + jobFile + ">");
				} catch (Exception e) {
					error("Can't parse job file <" + jobFile + ">", e);
					endProcessing();
				}
			}
		} else {
			try {
				ByteArrayOutputStream bos = new ByteArrayOutputStream();
				int b = 0;
				while ((b = System.in.read()) != -1) {
					bos.write(b);
				}
				builder = new Builder(bos.toString());
				verbose("Read job from stdin.");
			} catch (Exception e) {
				error("Can't read job from stdin.",e);
				endProcessing();
			}
		}
	}

	protected void runBes() {
		try {
			createActivity();
			if(verbose){
				String url = activityEpr.getAddress().getStringValue();
				verbose("Job submitted, job url=" + url);
			}
			updateBuilder();
		} catch (Exception e) {
			error("Could not run job.", e);
			endProcessing(1);
		}

		writeJobIDFile();
		endProcessing(0);
	}
	
	
	protected void updateBuilder(){
		builder.setProperty("activity-epr", activityEpr.toString());
		builder.setProperty("type", "job");
	}
	
	
	protected void createActivity(){

		CreateActivityDocument activityReqDoc = CreateActivityDocument.Factory.newInstance();
		CreateActivityType activityReq = activityReqDoc.addNewCreateActivity();
		ActivityDocumentType activityDoc = activityReq.addNewActivityDocument();
		try {
			activityDoc.setJobDefinition(builder.getJob().getJobDefinition());
		} catch (Exception e1) {
			error("Could not setup job definition.", e1);
			endProcessing(1);
		}
		TerminationTime tt = TerminationTime.Factory.newInstance();
		Calendar c = new GregorianCalendar();
		c.add(Calendar.MONTH, 2);
		tt.setCalendarValue(c);
		activityDoc.setTerminationTime(tt);
		// creating an activity
		CreateActivityResponseDocument response = null;
		try {
			response = factory.createActivity(activityReqDoc);
		} catch (NotAcceptingNewActivitiesFault e) {
			error("BES is not accepting new activities: ", e);
		} catch (InvalidRequestMessageFault e) {
			error("invalid request message: ", e);
		} catch (UnsupportedFeatureFault e) {
			error("Unsupported feature fault: ", e);
		}
		activityDocument = response.getCreateActivityResponse().getActivityDocument();
		
		if(response.getCreateActivityResponse()!=null)
			activityEpr = response.getCreateActivityResponse().getActivityIdentifier();
		// dump job properties for getting status
		if(activityDocument != null)dumpJobProperties();
		
	}
	
	

	protected String getStdOutName() {
		return getOption(OPT_STDOUT_LONG, OPT_STDOUT, "stdout");
	}

	protected String getStdErrName() {
		return getOption(OPT_STDERR_LONG, OPT_STDERR, "stderr");
	}

	protected void writeJobIDFile() {
		try {
			File dump = new File(output, ActivityUtils.extractRefParamResourceId(activityEpr)+ ".job");
			FileWriter fw = new FileWriter(dump);
			builder.setProperty("factory-epr", factory.getEPR().toString());
			builder.writeTo(fw);
			fw.close();
			verbose("Wrote job descriptor to " + dump.getAbsolutePath());
			lastJobDescriptor = dump.getAbsolutePath();
			message(dump.getAbsolutePath());
		} catch (Exception e) {
			error("Could not write job ID file.", e);
		}
	}


	protected void dumpJobProperties() {
		try {
			ActivityDocumentDocument1 ad = ActivityDocumentDocument1.Factory.newInstance();
			ad.setActivityDocument(activityDocument);
			String p = ad.toString();
			File dump = new File(output, ActivityUtils.extractRefParamResourceId(activityEpr)+ ".properties");
			FileWriter fw = new FileWriter(dump);
			fw.append(p);
			fw.close();
			message(dump.getAbsolutePath());
		} catch (Exception e) {
			error("Could not get job properties.", e);
		}
	}

	private static String lastJobDescriptor;

	public static String getLastJobDescriptor() {
		return lastJobDescriptor;
	}
	
	
	@Override
	public String getCommandGroup() {
		return BESUCCConstants.BES_COMMAND_GROUP;
	}
	

}
